import os
import json
import socket
from typing import Iterable
import redis
from .base import EwoksEventReader, EventType


class RedisEwoksEventReader(EwoksEventReader):
    def __init__(self, url: str, **_):
        client_name = f"ewoks:reader:{socket.gethostname()}:{os.getpid()}"
        self._proxy = redis.Redis.from_url(url, client_name=client_name)
        super().__init__()

    def wait_events(self, **kwargs) -> Iterable[EventType]:
        yield from self.poll_events(**kwargs)

    def get_events(self, job_id=None, **filters) -> Iterable[EventType]:
        is_equal_filter, post_filter = self.split_filter(**filters)

        if job_id:
            pattern = f"ewoks:{job_id}:*"
        else:
            pattern = "ewoks:*"
        keys = sorted(
            self._proxy.scan_iter(pattern), key=lambda x: int(x.decode().split(":")[-1])
        )
        for key in keys:
            event = self._proxy.hgetall(key)
            event = {k.decode(): json.loads(v) for k, v in event.items()}
            if not self.event_passes_filter(event, **post_filter) or any(
                event[k] != v for k, v in is_equal_filter.items()
            ):
                continue
            yield event



import os
import getpass
import json
from typing import Dict, Tuple, Any
from multiprocessing import get_context
from multiprocessing import get_all_start_methods

from click import Choice
from celery import Celery
from celery import bootsteps
from celery import concurrency
from celery.bin import worker
from celery.bin.base import CeleryOption

from .slurm import TaskPool as SlurmTaskPool
from .process import TaskPool as ProcessTaskPool

ALL_MP_CONTEXTS = list(get_all_start_methods())
DEFAULT_MP_CONTEXT = get_context()._name

concurrency.ALIASES["process"] = (
    f"{ProcessTaskPool.__module__}:{ProcessTaskPool.__name__}"
)
concurrency.ALIASES["slurm"] = f"{SlurmTaskPool.__module__}:{SlurmTaskPool.__name__}"
worker.WORKERS_POOL.choices = list(worker.WORKERS_POOL.choices) + ["process", "slurm"]


def add_options(app: Celery) -> None:
    _add_slurm_pool_options(app)
    _add_process_pool_options(app)
    app.steps["worker"].add(CustomWorkersBootStep)


class CustomWorkersBootStep(bootsteps.Step):
    def __init__(self, parent, **options):
        apply_worker_options(options)
        super().__init__(parent, **options)


def apply_worker_options(options):
    SlurmTaskPool.EXECUTOR_OPTIONS = _extract_slurm_options(options)
    ProcessTaskPool.EXECUTOR_OPTIONS = _extract_process_options(options)


def _add_slurm_pool_options(app: Celery) -> None:
    app.user_options["preload"].add(
        CeleryOption(
            ["--slurm-url"],
            required=False,
            default=os.environ.get("SLURM_URL"),
            help="SLURM REST URL",
            help_group="Slurm Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["--slurm-token"],
            required=False,
            default=os.environ.get("SLURM_TOKEN"),
            help="SLURM REST access token",
            help_group="Slurm Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["--slurm-user"],
            required=False,
            default=os.environ.get("SLURM_USER", getpass.getuser()),
            help="SLURM user name",
            help_group="Slurm Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["--slurm-log-directory"],
            required=False,
            help="Directory for SLURM to store the STDOUT and STDERR files",
            help_group="Slurm Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["--slurm-data-directory"],
            required=False,
            help="Directory for SLURM data transfer over files (TCP otherwise)",
            help_group="Slurm Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["--slurm-pre-script"],
            required=False,
            help="Script to be executes before each SLURM job (e.g. activate python environment)",
            help_group="Slurm Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["--slurm-post-script"],
            required=False,
            help="Script to be executes after each SLURM job",
            help_group="Slurm Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["-sp", "--slurm-parameter", "slurm_parameters"],
            required=False,
            multiple=True,
            help="SLURM job parameters (-sp NAME=VALUE).",
            help_group="Slurm Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["--slurm-python-cmd"],
            required=False,
            help="Python command (Default: python3)",
            help_group="Slurm Pool Options",
        )
    )


def _add_process_pool_options(app: Celery) -> None:
    app.user_options["preload"].add(
        CeleryOption(
            ["--process-context"],
            required=False,
            type=Choice(ALL_MP_CONTEXTS, case_sensitive=False),
            default=DEFAULT_MP_CONTEXT,
            show_default=True,
            help="Child process creation",
            help_group="Process Pool Options",
        )
    )
    app.user_options["preload"].add(
        CeleryOption(
            ["--process-no-precreate"],
            required=False,
            default=False,
            help="Child processes not created on startup",
            help_group="Process Pool Options",
        )
    )


SLURM_NAME_MAP = {
    "slurm_url": "url",
    "slurm_user": "user_name",
    "slurm_token": "token",
    "slurm_log_directory": "log_directory",
    "slurm_data_directory": "data_directory",
    "slurm_pre_script": "pre_script",
    "slurm_post_script": "post_script",
    "slurm_parameters": "parameters",
    "slurm_python_cmd": "python_cmd",
}


def _extract_slurm_options(options: Dict) -> dict:
    slurm_options = {
        name: options.get(option) for option, name in SLURM_NAME_MAP.items()
    }
    parameters = slurm_options.pop("parameters", None)
    if parameters:
        slurm_options["parameters"] = dict(
            _parse_slurm_parameter(p) for p in parameters
        )
    return slurm_options


def _parse_slurm_parameter(parameter: str) -> Tuple[str, Any]:
    name, _, value = parameter.partition("=")
    return name, _parse_value(value)


def _parse_value(value: str) -> Any:
    try:
        return json.loads(value)
    except Exception:
        return value


PROCESS_NAME_MAP = {
    "process_context": "context",
    "process_no_precreate": "precreate",
}


def _extract_process_options(options: Dict) -> dict:
    process_options = {
        name: options.get(option) for option, name in PROCESS_NAME_MAP.items()
    }
    process_options["precreate"] = not process_options["precreate"]
    return process_options



import pytest
from ewokscore import Task
from ewoksutils.import_utils import qualname
from ewoks import execute_graph
from .utils import has_redis


class AddNumbers(Task, input_names=["a", "b"], output_names=["sum"]):
    def run(self):
        self.outputs.sum = self.inputs.a + self.inputs.b


def generate_graph():
    return {
        "graph": {"id": "test"},
        "nodes": [
            {
                "id": "task",
                "task_identifier": qualname(AddNumbers),
                "task_type": "class",
            }
        ],
    }


@pytest.mark.skipif(not has_redis(), reason="redis-server not installed")
@pytest.mark.parametrize("scheme", ("nexus", "json"))
def test_redis(scheme, redis_ewoks_events, tmpdir):
    handlers, reader = redis_ewoks_events
    assert_feedback(scheme, handlers, reader, tmpdir)


@pytest.mark.parametrize("scheme", ("nexus", "json"))
def test_sqlite3(scheme, sqlite3_ewoks_events, tmpdir):
    handlers, reader = sqlite3_ewoks_events
    assert_feedback(scheme, handlers, reader, tmpdir)


def assert_feedback(scheme, handlers, reader, tmpdir):
    execinfo = {"handlers": handlers}
    graph = generate_graph()

    return_value = execute_graph(
        graph,
        execinfo=execinfo,
        varinfo={"root_uri": str(tmpdir), "scheme": scheme},
        inputs=[
            {"id": "task", "name": "a", "value": 1},
            {"id": "task", "name": "b", "value": 2},
        ],
        outputs=[{"id": "task", "name": "sum"}],
    )
    assert return_value == {"sum": 3}

    assert len(list(reader.get_events_with_variables())) == 6

    evts = list(reader.get_events_with_variables(node_id="task", type="start"))
    assert len(evts) == 1

    event_values = evts[0]["outputs"]
    assert event_values.get_variable_values() == {"sum": 3}



"""Remote worker pool managed by Celery
"""

import os
from billiard.exceptions import Terminated

try:
    from gevent.monkey import is_anything_patched
    from gevent.monkey import is_module_patched
except ImportError:
    pass
else:
    if is_anything_patched() and not is_module_patched("threading"):
        # Make Celery use `celery.backends.asynchronous.Drainer`
        # instead of `celery.backends.asynchronous.geventDrainer`.
        # The later causes CTRL-C to not be raised and other things
        # like Bliss scans to hang when calling `AsyncResult.get`.
        from kombu.utils import compat

        compat._environment = "default"

        # The real solution is to patch threads.

from celery.exceptions import TaskRevokedError as CancelledError

CancelledErrors = CancelledError, Terminated
from .tasks import *  # noqa F403
from .utils import *  # noqa F403
from .tasks import execute_graph as submit  # noqa F401
from .tasks import execute_test_graph as submit_test  # noqa F401

# For clients (workers need it in the environment before stating the python process)
os.environ.setdefault("CELERY_LOADER", "ewoksjob.config.EwoksLoader")



import logging
from typing import Dict, List, Optional, Set
from celery import current_app
from celery.result import AsyncResult

__all__ = [
    "get_future",
    "cancel",
    "get_result",
    "get_not_finished_task_ids",
    "get_not_finished_futures",
    "get_workers",
]


logger = logging.getLogger(__name__)


def get_future(task_id) -> AsyncResult:
    return AsyncResult(task_id)


def cancel(task_id):
    future = get_future(task_id)
    if future is not None:
        future.revoke(terminate=True)


def get_result(task_id, **kwargs):
    kwargs.setdefault("interval", 0.1)
    future = AsyncResult(task_id)
    if future is not None:
        return future.get(**kwargs)


def get_not_finished_task_ids():
    inspect = current_app.control.inspect()
    task_ids = list()

    workers = inspect.active()  # running
    if workers is None:
        logger.warning("No Celery workers were detected")
        workers = dict()
    for tasks in workers.values():
        for task in tasks:
            task_ids.append(task["id"])

    workers = inspect.scheduled()  # pending
    if workers is None:
        workers = dict()
    for tasks in workers.values():
        for task in tasks:
            task_ids.append(task["id"])

    return task_ids


def get_not_finished_futures() -> List[AsyncResult]:
    lst = [get_future(task_id) for task_id in get_not_finished_task_ids()]
    return [future for future in lst if future is not None]


def get_workers() -> List[str]:
    queues_dict: Optional[Dict[str, List[dict]]] = (
        current_app.control.inspect().active_queues()
    )
    if queues_dict is None:
        return list()

    workers: Set[str] = set()
    for queue_infos in queues_dict.values():
        for queue_info in queue_infos:
            workers.add(queue_info["name"])

    return list(workers)



from ewokscore.tests.examples.graphs import get_graph
from ..client import celery
from ..client import local
from .utils import get_result


def test_submit(ewoks_worker):
    assert_submit(celery)
    assert_submit_test(celery)


def test_submit_local(local_ewoks_worker):
    assert_submit(local)
    assert_submit_test(local)


def assert_submit(mod):
    graph, expected = get_graph("acyclic1")
    expected = expected["task6"]
    future1 = mod.submit(args=(graph,))
    future2 = mod.get_future(future1.task_id)
    results = get_result(future1, timeout=10)
    assert results == expected
    results = get_result(future2, timeout=0)
    assert results == expected


def assert_submit_test(mod):
    future1 = mod.submit_test()
    future2 = mod.get_future(future1.task_id)
    results = get_result(future1, timeout=10)
    assert results == {"return_value": True}
    results = get_result(future2, timeout=0)
    assert results == {"return_value": True}



import time

import pytest
from ewokscore.tests.examples.graphs import get_graph

try:
    import pyicat_plus  # noqa F401

    ICAT_ERROR_MSG = "The message queue URL's are missing"
except ImportError:
    ICAT_ERROR_MSG = "requires pyicat-plus"

from ..client import celery
from ..client import local
from .utils import get_result


def test_submit(ewoks_worker):
    assert_submit(celery)


def test_submit_local(local_ewoks_worker):
    assert_submit(local)


def assert_submit(mod):
    graph, expected = get_graph("acyclic1")
    expected = expected["task6"]
    kwargs = {
        "upload_parameters": {
            "metadata_urls": list(),
            "beamline": "id00",
            "proposal": f"id00{time.strftime('%y%m')}",
            "dataset": "testdataset",
            "path": "/path/to/localed/dataset",
            "metadata": {"Sample_name": "test"},
            "raw": "/path/to/raw/dataset",
        }
    }
    future1 = mod.submit(args=(graph,), kwargs=kwargs)
    with pytest.raises(RuntimeError, match=ICAT_ERROR_MSG):
        get_result(future1, timeout=10)



import pytest
from ..client import celery
from ..client import local
from .utils import wait_not_finished
from .utils import has_redis


@pytest.mark.skip(reason="working in manual testing")
@pytest.mark.skipif(
    not has_redis(), reason="memory and sqlite do not support task monitoring"
)
def test_futures(ewoks_worker):
    assert_futures(celery)


def test_futures_local(local_ewoks_worker):
    assert_futures(local)


def assert_futures(mod):
    seconds = 2
    timeout = seconds * 4
    future1 = mod.submit_test(seconds)
    future2 = mod.submit_test(seconds)
    tasks1 = {future1.task_id, future2.task_id}
    wait_not_finished(mod, tasks1, timeout=timeout)
    tasks2 = {future.task_id for future in mod.get_not_finished_futures()}
    assert tasks1 == tasks2
    wait_not_finished(mod, set(), timeout=timeout)



"""Celery ewoks application executed on the worker side."""

from functools import wraps
from typing import Dict, List, Union, Callable, Any

import celery
import ewoks
from ewokscore import task_discovery

from ..worker.options import add_options
from ..worker.executor import get_executor
from .errors import replace_exception_for_client

app = celery.Celery("ewoks")
add_options(app)


def _ensure_ewoks_job_id(celery_task: Callable) -> Callable:
    """Use celery task ID as ewoks job ID when ewoks job ID is not provided"""

    @wraps(celery_task)
    def new_celery_task(self, *args, **kwargs):
        execinfo = kwargs.setdefault("execinfo", dict())
        if not execinfo.get("job_id"):
            execinfo["job_id"] = self.request.id
        return celery_task(self, *args, **kwargs)

    return new_celery_task


def _task_wrapper(celery_task: Callable) -> Callable:
    """Wraps all celery tasks in order to

    * convert exceptions that the client is not expected to have
    * execute with a worker specific executor (e.g. redirect the
      ewoks task to another job scheduler)
    """

    @wraps(celery_task)
    def new_celery_task(*args, **kwargs) -> Any:
        with replace_exception_for_client():
            executor = get_executor()

            if executor is None:
                return celery_task(*args, **kwargs)

            # Remove all references to ewoksjob
            ewoks_task = _get_ewoks_task_from_celery_task(celery_task)
            if _celery_task_is_bound(celery_task):
                args = args[1:]  # remove `self`

            return executor(ewoks_task, *args, **kwargs)

    return new_celery_task


@app.task(bind=True)
@_ensure_ewoks_job_id
@_task_wrapper
def execute_graph(self, *args, **kwargs) -> Dict:
    return ewoks.execute_graph(*args, **kwargs)


@app.task()
@_task_wrapper
def convert_graph(*args, **kwargs) -> Union[str, dict]:
    return ewoks.convert_graph(*args, **kwargs)


@app.task()
@_task_wrapper
def discover_tasks_from_modules(*args, **kwargs) -> List[dict]:
    return task_discovery.discover_tasks_from_modules(*args, **kwargs)


@app.task()
@_task_wrapper
def discover_all_tasks(*args, **kwargs) -> List[dict]:
    return task_discovery.discover_all_tasks(*args, **kwargs)


_TASK_MAPPING: Dict[Callable, Callable] = {
    "execute_graph": ewoks.execute_graph,
    "convert_graph": ewoks.convert_graph,
    "discover_tasks_from_modules": task_discovery.discover_tasks_from_modules,
    "discover_all_tasks": task_discovery.discover_all_tasks,
}

_BOUND_TASKS = {"execute_graph"}


def _get_ewoks_task_from_celery_task(celery_task: Callable) -> Callable:
    return _TASK_MAPPING[celery_task.__name__]


def _celery_task_is_bound(celery_task: Callable) -> Callable:
    return celery_task.__name__ in _BOUND_TASKS



import sys
from contextlib import contextmanager
from typing import Mapping, Optional, Tuple
from uuid import uuid4
import multiprocessing
import weakref
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import Future

try:
    from pyslurmutils.concurrent.futures import SlurmRestExecutor
    from pyslurmutils.client.job_io.file_io import Future as SlurmFuture
except ImportError:
    SlurmRestExecutor = None


__all__ = ["get_active_pool", "pool_context"]


_EWOKS_WORKER_POOL = None


def get_active_pool(raise_on_missing: Optional[bool] = True):
    if raise_on_missing and _EWOKS_WORKER_POOL is None:
        raise RuntimeError("No worker pool is available")
    return _EWOKS_WORKER_POOL


@contextmanager
def pool_context(*args, **kwargs):
    global _EWOKS_WORKER_POOL
    if _EWOKS_WORKER_POOL is None:
        with _LocalPool(*args, **kwargs) as pool_obj:
            _EWOKS_WORKER_POOL = pool_obj
            try:
                yield pool_obj
            finally:
                _EWOKS_WORKER_POOL = None
    else:
        yield _EWOKS_WORKER_POOL


class _LocalPool:
    def __init__(
        self,
        *args,
        pool_type: Optional[str] = None,
        context: Optional[str] = None,
        **kwargs,
    ) -> None:
        if pool_type is None:
            pool_type = "process"
        if context is None:
            context = "spawn"
        if pool_type == "process":
            if context:
                if sys.version_info >= (3, 7):
                    kwargs["mp_context"] = multiprocessing.get_context(context)
                else:
                    multiprocessing.set_start_method(context, force=True)
            self._executor = ProcessPoolExecutor(*args, **kwargs)
        elif pool_type == "thread":
            self._executor = ThreadPoolExecutor(*args, **kwargs)
        elif pool_type == "slurm":
            if SlurmRestExecutor is None:
                raise RuntimeError("requires pyslurmutils")
            self._executor = SlurmRestExecutor(*args, **kwargs)
        else:
            raise ValueError(f"Unknown pool type '{pool_type}'")
        self._pool_type = pool_type
        self._tasks = weakref.WeakValueDictionary()

    def __enter__(self):
        self._executor.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._executor.__exit__(exc_type, exc_val, exc_tb)

    def shutdown(self, **kw):
        return self._executor.shutdown(**kw)

    @property
    def pool_type(self):
        return self._pool_type

    def submit(
        self,
        func,
        task_id=None,
        args: Optional[Tuple] = tuple(),
        kwargs: Optional[Mapping] = None,
    ) -> Future:
        """Like celery.send_task"""
        if kwargs is None:
            kwargs = dict()
        future = self._executor.submit(func, *args, **kwargs)
        self._patch_future(future, task_id)
        self._tasks[future.task_id] = future
        return future

    def generate_task_id(self, task_id=None):
        if self.pool_type == "slurm":
            return task_id
        if task_id is None:
            task_id = str(uuid4())
            while task_id in self._tasks:
                task_id = str(uuid4())
            return task_id
        if task_id in self._tasks:
            raise RuntimeError(f"Job '{task_id}' already exists")
        return task_id

    def get_future(self, task_id):
        future = self._tasks.get(task_id)
        if future is None:
            return self._get_dummy_future(task_id)
        return future

    def _get_dummy_future(self, task_id):
        if self.pool_type == "slurm":
            future = SlurmFuture(0, "", self._executor.client.io_handler)
        else:
            future = Future()
        self._patch_future(future, self.generate_task_id(task_id))
        return future

    def get_not_finished_task_ids(self) -> list:
        """Get all task ID's that are not finished"""
        return [task_id for task_id, future in self._tasks.items() if not future.done()]

    def _patch_future(self, future, task_id):
        if not task_id and self.pool_type == "slurm":
            future.task_id = str(future.job_id)
        else:
            future.task_id = task_id
        # Warning: this causes the future to never be garbage collected
        # future.get = future.result



def result(value=False, filename=None):
    if filename:
        with open(filename, "w"):
            pass
    return value is None


def dummy_workflow():
    return {
        "graph": {"id": "sleepgraph", "schema_version": "1.0"},
        "nodes": [
            {
                "id": "sleep",
                "task_type": "method",
                "task_identifier": "time.sleep",
                "default_inputs": [{"name": 0, "value": 0}],
            },
            {
                "id": "result",
                "task_type": "method",
                "task_identifier": "ewoksjob.client.dummy_workflow.result",
                "default_inputs": [
                    {"name": "value", "value": None},
                    {"name": "filename", "value": None},
                ],
            },
        ],
        "links": [
            {
                "source": "sleep",
                "target": "result",
                "data_mapping": [
                    {"source_output": "return_value", "target_input": "value"}
                ],
            },
        ],
    }



import pytest
from celery.exceptions import TimeoutError as CeleryTimeoutError
from concurrent.futures import TimeoutError as localTimeoutError
from ..client import celery
from ..client import local
from .utils import get_result


def test_task_discovery(ewoks_worker):
    future = celery.get_future("abc")
    assert future.status == "PENDING"
    with pytest.raises(CeleryTimeoutError):
        future.get(timeout=1e-8)


def test_task_discovery_local(local_ewoks_worker):
    future = local.get_future("abc")
    assert not future.running()
    with pytest.raises(localTimeoutError):
        get_result(future, timeout=0)



import os
import sys
import types
import logging
import importlib
from pathlib import Path
from typing import Tuple
from urllib.parse import urlparse, ParseResult

from celery.loaders.default import Loader
from celery import Celery

try:
    from blissdata.beacon.files import read_config as bliss_read_config
except ImportError:
    bliss_read_config = None

logger = logging.getLogger(__name__)


class EwoksLoader(Loader):
    """Celery loader based on a configuration URI: python file, python module, yaml file, Beacon URL.

    Requires the environment variable CELERY_LOADER=ewoksjob.config.EwoksLoader
    """

    def __init__(self, app: Celery) -> None:
        self.app = app
        super().__init__(app)

    def read_configuration(self) -> dict:
        if "" not in sys.path:
            # This happens when the python process was launched
            # through a python console script
            sys.path.append("")

        cfg_uri = get_cfg_uri()
        if not cfg_uri:
            # Load from the module "celeryconfig" if available
            return super().read_configuration()

        try:
            file_type = get_cfg_type(cfg_uri)
            logger.info(f"Loading celery configuration '{cfg_uri}' (type: {file_type})")
            config = read_configuration(cfg_uri=cfg_uri)
        except Exception:
            raise RuntimeError(
                f"Cannot load celery configuration from '{cfg_uri}' (type: {file_type})"
            )
        return config


def get_cfg_uri() -> str:
    """Returns the celery configuration URI based on environment variables."""
    cfg_uri = os.environ.get("EWOKS_CONFIG_URI")
    if cfg_uri:
        return cfg_uri
    cfg_uri = os.environ.get("CELERY_CONFIG_URI")  # deprecate?
    if cfg_uri:
        return cfg_uri
    beacon_host = os.environ.get("BEACON_HOST", None)
    if beacon_host:
        return "beacon:///ewoks/config.yml"
    cfg_uri = os.environ.get("CELERY_CONFIG_MODULE")
    # the CLI option --config sets this environment variable
    if cfg_uri:
        return cfg_uri
    return ""


def get_cfg_type(cfg_uri: str) -> str:
    presult = _parse_url(cfg_uri)
    if presult.scheme == "beacon":
        return "yaml"
    if presult.scheme in ("file", ""):
        cfg_uri = _url_to_filename(presult)
        ext = os.path.splitext(cfg_uri)[-1]
        if ext in (".yaml", ".yml"):
            return "yaml"
        return "python"
    return ""


def read_configuration(cfg_uri: str) -> dict:
    """Different types of URI's are supported:

    - Python module:
       - myproject.config
    - Python file:
       - /tmp/ewoks/config.py
    - Yaml file:
       - /tmp/ewoks/config.yml
    - Beacon yaml file:
       - beacon:///ewoks/config.yml  (this requires the BEACON_HOST environment variable)
       - beacon://id22:25000/ewoks/config.yml
    """
    file_type = get_cfg_type(cfg_uri)
    if file_type == "yaml":
        config = _read_yaml_config(cfg_uri)
    elif file_type == "python":
        config = _read_py_config(cfg_uri)
    else:
        raise ValueError(f"Configuration URL '{cfg_uri}' is not supported")
    if "celery" in config:
        config = config["celery"]
    elif "CELERY" in config:
        config = config["CELERY"]
    return config


def _parse_url(url: str) -> ParseResult:
    presult = urlparse(url)
    if presult.scheme == "beacon":
        # beacon:///path/to/file.yml
        # beacon://id00:25000/path/to/file.yml
        return presult
    elif presult.scheme in ("file", ""):
        # /path/to/file.yaml
        # file:///path/to/file.yaml
        return presult
    elif sys.platform == "win32" and len(presult.scheme) == 1:
        # c:\\path\\to\\file.yaml
        return urlparse(f"file://{url}")
    else:
        return presult


def _url_to_filename(presult: ParseResult) -> str:
    if presult.netloc and presult.path:
        # urlparse("file://c:/a/b")
        return presult.netloc + presult.path
    elif presult.netloc:
        # urlparse("file://c:\\a\\b")
        return presult.netloc
    else:
        # urlparse("file:///a/b")
        return presult.path


def _read_yaml_config(resource: str) -> dict:
    if bliss_read_config is None:
        raise RuntimeError(
            f"Cannot get celery configuration '{resource}' from Beacon: blissdata is not installed"
        )
    return bliss_read_config(resource)


def _read_py_config(cfg_uri: str) -> dict:
    """Warning: this is not thread-safe and it has side-effects during execution"""
    presult = _parse_url(cfg_uri)
    sys_path, module = _get_config_module(_url_to_filename(presult))
    keep_sys_path = list(sys.path)
    sys.path.insert(0, sys_path)
    try:
        config = vars(importlib.import_module(module))
        config = {
            k: v
            for k, v in config.items()
            if not k.startswith("_") and not isinstance(v, types.ModuleType)
        }
        return config
    finally:
        sys.path = keep_sys_path


def _get_config_module(cfg_uri: str) -> Tuple[str, str]:
    path = Path(cfg_uri)
    if path.is_file():
        parent = str(path.parent.resolve())
        return parent, path.stem
    return os.getcwd(), cfg_uri



import os
import pytest
from ..config import read_configuration

EXPECTED = {
    "broker_url": "redis://localhost:6379/3",
    "result_backend": "redis://localhost:6379/4",
    "result_serializer": "pickle",
    "accept_content": ["application/json", "application/x-python-serialize"],
    "result_expires": 600,
    "task_remote_tracebacks": True,
    "enable_utc": False,
}


def test_pyfile_config(py_config: str):
    assert read_configuration(py_config) == EXPECTED
    assert read_configuration(f"file://{py_config}") == EXPECTED


def test_pymodule_config(py_config: str):
    keep = os.getcwd()
    module = os.path.splitext(os.path.basename(py_config))[0]
    os.chdir(os.path.dirname(py_config))
    try:
        assert read_configuration(module) == EXPECTED
    finally:
        os.chdir(keep)


def test_yaml_config(yaml_config: str):
    assert read_configuration(yaml_config) == EXPECTED
    assert read_configuration(f"file://{yaml_config}") == EXPECTED


def test_beacon_config(beacon_config: str):
    assert read_configuration(beacon_config) == EXPECTED


@pytest.fixture
def py_config(tmpdir) -> str:
    filename = str(tmpdir / "celeryconfig.py")
    lines = [
        "broker_url = 'redis://localhost:6379/3'\n",
        "result_backend = 'redis://localhost:6379/4'\n",
        "result_serializer = 'pickle'\n",
        "accept_content = ['application/json', 'application/x-python-serialize']\n",
        "result_expires = 600\n",
        "task_remote_tracebacks = True\n",
        "enable_utc = False\n",
    ]
    with open(filename, "w") as f:
        f.writelines(lines)
    return filename


@pytest.fixture
def yaml_config(tmpdir) -> str:
    filename = str(tmpdir / "ewoks.yaml")
    lines = [
        "celery:\n",
        "  broker_url: 'redis://localhost:6379/3'\n",
        "  result_backend: 'redis://localhost:6379/4'\n",
        "  result_serializer: 'pickle'\n",
        "  accept_content: ['application/json', 'application/x-python-serialize']\n",
        "  result_expires: 600\n",
        "  task_remote_tracebacks: True\n",
        "  enable_utc: False\n",
    ]
    with open(filename, "w") as f:
        f.writelines(lines)
    return filename


@pytest.fixture
def beacon_config(mocker) -> str:
    url = "beacon://localhost:1234/config.yml"
    client = mocker.patch("ewoksjob.config.bliss_read_config")

    def read_config(_url):
        if _url == url:
            return EXPECTED

    client.side_effect = read_config
    return url



import sqlite3
from typing import Iterator
from ewoksutils.event_utils import FIELD_TYPES
from ewoksutils import sqlite3_utils
from .base import EwoksEventReader, EventType


class Sqlite3EwoksEventReader(EwoksEventReader):
    def __init__(self, uri: str, **_) -> None:
        super().__init__()
        self._uri = uri
        self.__connection = None
        self.__sql_types = sqlite3_utils.python_to_sql_types(FIELD_TYPES)

    def close(self):
        if self.__connection is not None:
            self.__connection.close()
            self.__connection = None
        super().close()

    @property
    def _connection(self):
        if self.__connection is None:
            self.__connection = sqlite3.connect(self._uri, uri=True)
        return self.__connection

    def wait_events(self, **kwargs) -> Iterator[EventType]:
        yield from self.poll_events(**kwargs)

    def get_events(self, **filters) -> Iterator[EventType]:
        yield from sqlite3_utils.select(
            self._connection,
            "ewoks_events",
            field_types=FIELD_TYPES,
            sql_types=self.__sql_types,
            **filters
        )



import os
import sys
import functools
import celery


def _patch_eventlet():
    import eventlet.debug

    eventlet.monkey_patch()
    blockdetect = float(os.environ.get("EVENTLET_NOBLOCK", 0))
    if blockdetect:
        eventlet.debug.hub_blocking_detection(blockdetect, blockdetect)


def _patch_gevent():
    import gevent.monkey
    import gevent.signal

    gevent.monkey.patch_all()


_PATCHES = {
    "eventlet": _patch_eventlet,
    "gevent": _patch_gevent,
    "slurm": _patch_gevent,
}

celery.maybe_patch_concurrency = functools.partial(
    celery.maybe_patch_concurrency, patches=_PATCHES
)


def main(argv=None) -> None:
    if argv is None:
        argv = sys.argv
    if "worker" in argv and "-A" not in argv:
        argv = argv[:1] + ["-A", "ewoksjob.apps.ewoks"] + argv[1:]
    if "monitor" in argv:
        argv[argv.index("monitor")] = "flower"
    sys.argv = argv
    os.environ.setdefault("CELERY_LOADER", "ewoksjob.config.EwoksLoader")

    from celery.__main__ import main as _main

    sys.exit(_main())


if __name__ == "__main__":
    main()



from typing import List, Optional
from concurrent.futures import Future
from .pool import get_active_pool


__all__ = [
    "get_future",
    "cancel",
    "get_result",
    "get_not_finished_task_ids",
    "get_not_finished_futures",
]


def get_future(task_id) -> Optional[Future]:
    pool = get_active_pool()
    return pool.get_future(task_id)


def cancel(task_id):
    """The current implementation does not allow cancelling running tasks"""
    future = get_future(task_id)
    if future is not None:
        future.cancel()


def get_result(task_id, **kwargs):
    future = get_future(task_id)
    if future is not None:
        return future.result(**kwargs)


def get_not_finished_task_ids() -> list:
    """Get all task ID's that are not finished"""
    pool = get_active_pool()
    return pool.get_not_finished_task_ids()


def get_not_finished_futures() -> List[Future]:
    """Get all futures that are not finished"""
    lst = [get_future(task_id) for task_id in get_not_finished_task_ids()]
    return [future for future in lst if future is not None]



from .base import *  # noqa F401
from .sqlite3 import Sqlite3EwoksEventReader  # noqa F401

try:
    from .redis import RedisEwoksEventReader  # noqa F401
except ImportError:
    pass


def instantiate_reader(url: str, **kw) -> EwoksEventReader:  # noqa F405
    s = url.lower()
    if any(s.startswith(scheme) for scheme in ("redis:", "rediss:", "unix:")):
        return RedisEwoksEventReader(url, **kw)
    elif s.startswith("file:"):
        return Sqlite3EwoksEventReader(url, **kw)
    else:
        raise ValueError(f"unknown scheme for '{url}'")



from ewokscore.tests.examples.graphs import get_graph
from ..client import celery
from ..client import local
from .utils import get_result


def test_submit(ewoks_worker, tmpdir):
    assert_submit(celery, tmpdir)
    assert_submit_test(celery, tmpdir)


def test_submit_local(local_ewoks_worker, tmpdir):
    assert_submit(local, tmpdir)
    assert_submit_test(local, tmpdir)


def assert_submit(mod, tmpdir):
    graph, expected = get_graph("acyclic1")
    expected = expected["task6"]
    filename = tmpdir / "test.json"
    args = (graph,)
    kwargs = {"save_options": {"indent": 2}, "convert_destination": str(filename)}
    future1 = mod.submit(args=args, kwargs=kwargs)
    future2 = mod.get_future(future1.task_id)
    results = get_result(future1, timeout=10)
    assert results == expected
    results = get_result(future2, timeout=0)
    assert results == expected
    assert filename.exists()


def assert_submit_test(mod, tmpdir):
    filename = tmpdir / "test.json"
    kwargs = {"save_options": {"indent": 2}, "convert_destination": str(filename)}
    future1 = mod.submit_test(kwargs=kwargs)
    future2 = mod.get_future(future1.task_id)
    results = get_result(future1, timeout=10)
    assert results == {"return_value": True}
    results = get_result(future2, timeout=0)
    assert results == {"return_value": True}
    assert filename.exists()



"""Pool that redirects tasks to a Slurm cluster."""

import weakref
import logging
from functools import wraps
from typing import Callable

from celery.concurrency.gevent import TaskPool as _TaskPool

try:
    from pyslurmutils.client import SlurmPythonJobRestClient
except ImportError:
    SlurmPythonJobRestClient = None

from .executor import set_executor_getter, ExecutorType


__all__ = ("TaskPool",)

logger = logging.getLogger(__name__)


class TaskPool(_TaskPool):
    """SLURM Task Pool."""

    EXECUTOR_OPTIONS = dict()

    def __init__(self, *args, **kwargs):
        if SlurmPythonJobRestClient is None:
            raise RuntimeError("requires pyslurmutils")
        super().__init__(*args, **kwargs)
        self._create_slurm_client()

    def restart(self):
        self._remove_slurm_client()
        self._create_slurm_client()

    def on_stop(self):
        self._remove_slurm_client()
        super().on_stop()

    def terminate_job(self, pid, signal=None):
        print("TODO: support job cancelling for the slurm pool")

    def _create_slurm_client(self):
        self._slurm_client = SlurmPythonJobRestClient(
            max_workers=self.limit, **self.EXECUTOR_OPTIONS
        )
        _set_slurm_client(self._slurm_client)

    def _remove_slurm_client(self):
        self._slurm_client.cleanup()
        self._slurm_client = None


_SLURM_CLIENT = None


def _set_slurm_client(slurm_client):
    global _SLURM_CLIENT
    _SLURM_CLIENT = weakref.proxy(slurm_client)
    set_executor_getter(_get_executor)


def _get_executor() -> ExecutorType:
    try:
        spawn = _SLURM_CLIENT.spawn
    except (AttributeError, ReferenceError):
        # TaskPool is not instantiated
        return
    return _slurm_execute(spawn)


def _slurm_execute(spawn: Callable) -> Callable:
    """Instead of executing the celery task, forward the ewoks task to Slurm."""

    @wraps(spawn)
    def executor(ewoks_task: Callable, *args, **kwargs):
        future = spawn(ewoks_task, args=args, kwargs=kwargs)
        try:
            return future.result()
        except BaseException:
            future.client.log_stdout_stderr(
                future.job_id, logger=logger, level=logging.ERROR
            )
            raise
        else:
            future.client.log_stdout_stderr(
                future.job_id, logger=logger, level=logging.INFO
            )
        finally:
            try:
                status = future.job_status()
                logger.info("Slurm job %s, %s", future.job_id, status)
                if future.job_status() not in ("COMPLETED", "CANCELLED", "FAILED"):
                    future.cancel_job()
            finally:
                future.cleanup_job()

    return executor



from ..client import celery
from ..client import local
from .utils import get_result


def test_convert(ewoks_worker, tmpdir):
    assert_convert(celery, tmpdir)


def test_convert_local(local_ewoks_worker, tmpdir):
    assert_convert(local, tmpdir)


def assert_convert(mod, tmpdir):
    filename = tmpdir / "test.json"
    args = {"graph": {"id": "testgraph", "schema_version": "1.0"}}, str(filename)
    kwargs = {"save_options": {"indent": 2}}
    future = mod.convert_workflow(args=args, kwargs=kwargs)
    results = get_result(future, timeout=10)
    assert results == str(filename) or results is None  # TODO: None is deprecated
    assert filename.exists()



from ..client import celery
from ..client import local
from .utils import get_result


def test_submit(ewoks_worker):
    assert_submit(celery)
    assert_submit(celery, "ewokscore")
    assert_submit(celery)


def test_submit_local(local_ewoks_worker):
    assert_submit(local)
    assert_submit(local, "ewokscore")
    assert_submit(local)


def assert_submit(mod, *modules):
    if modules:
        future1 = mod.discover_tasks_from_modules(args=modules)
    else:
        future1 = mod.discover_all_tasks()
    future2 = mod.get_future(future1.task_id)
    results = get_result(future1, timeout=10)
    assert results
    results = get_result(future2, timeout=0)
    assert results



import pytest
from ..client import celery
from ..client import local
from .utils import wait_not_finished


def test_normal(ewoks_worker, tmpdir):
    assert_normal(celery, tmpdir)


def test_normal_local(local_ewoks_worker, tmpdir):
    assert_normal(local, tmpdir)


def test_cancel(ewoks_worker, tmpdir, skip_if_gevent):
    assert_cancel(celery, tmpdir)


def test_cancel_local(local_ewoks_worker, tmpdir, skip_if_gevent):
    assert_cancel(local, tmpdir)


def assert_normal(mod, tmpdir):
    seconds = 5
    timeout = 10
    filename = tmpdir / "finished.smf"
    future = mod.submit_test(seconds=seconds, filename=str(filename))
    wait_not_finished(mod, {future.task_id}, timeout=timeout)
    results = mod.get_result(future.task_id, timeout=timeout)
    assert results == {"return_value": True}
    assert filename.exists()
    wait_not_finished(mod, set(), timeout=timeout)


def assert_cancel(mod, tmpdir):
    seconds = 10
    timeout = seconds * 2
    filename = tmpdir / "finished.smf"
    future = mod.submit_test(seconds=seconds, filename=str(filename))

    if mod is local:
        # The current implementation does not allow cancelling running tasks
        mod.cancel(future.task_id)
        try:
            results = mod.get_result(future.task_id, timeout=timeout)
        except mod.CancelledErrors:
            assert not filename.exists()
        else:
            assert results == {"return_value": True}
            assert filename.exists()
            pytest.xfail(f"{mod.__name__} ran until completion")
    else:
        wait_not_finished(mod, {future.task_id}, timeout=timeout)
        mod.cancel(future.task_id)
        try:
            results = mod.get_result(future.task_id, timeout=timeout)
        except mod.CancelledErrors:
            assert not filename.exists()
        else:
            assert results == {"return_value": True}
            assert filename.exists()
            pytest.xfail(f"{mod.__name__} ran until completion")

    # TODO: checking the futures is not enough, check with celery
    wait_not_finished(mod, set(), timeout=timeout)



import gc
import os
import pytest

from ewokscore import events
from ewoksjob.events.readers import instantiate_reader
from ewoksjob.worker import options as worker_options

from .utils import has_redis
from ..client import local

try:
    from pyslurmutils.tests.conftest import slurm_data_directory  # noqa F401
    from pyslurmutils.tests.conftest import slurm_log_directory  # noqa F401
    from pyslurmutils.tests.conftest import data_directory  # noqa F401
    from pyslurmutils.tests.conftest import log_directory  # noqa F401
    from pyslurmutils.tests.conftest import slurm_env  # noqa F401
    from pyslurmutils.tests.conftest import slurm_config
except ImportError:

    @pytest.fixture(scope="session")
    def slurm_config() -> None:
        pytest.skip("requires pyslurmutils")


if has_redis():
    import redis

    @pytest.fixture(scope="session")
    def celery_config(redis_proc):
        url = f"redis://{redis_proc.host}:{redis_proc.port}"
        # celery -A ewoksjob.apps.ewoks --broker={url}/0 --result-backend={url}/1 inspect stats -t 5
        return {
            "broker_url": f"{url}/0",
            "result_backend": f"{url}/1",
            "result_serializer": "pickle",
            "accept_content": ["application/json", "application/x-python-serialize"],
            "task_remote_tracebacks": True,
            "enable_utc": False,
        }

else:

    @pytest.fixture(scope="session")
    def celery_config(tmpdir_factory):
        tmpdir = tmpdir_factory.mktemp("celery")
        return {
            "broker_url": "memory://",
            # "broker_url": f"sqla+sqlite:///{tmpdir / 'celery.db'}",
            "result_backend": f"db+sqlite:///{tmpdir / 'celery_results.db'}",
            "result_serializer": "pickle",
            "accept_content": ["application/json", "application/x-python-serialize"],
            "task_remote_tracebacks": True,
            "enable_utc": False,
        }


@pytest.fixture(scope="session")
def celery_includes():
    return ("ewoksjob.apps.ewoks",)


@pytest.fixture(scope="session")
def celery_worker_parameters(slurm_config):
    if _use_slurm_pool():
        rmap = {v: k for k, v in worker_options.SLURM_NAME_MAP.items()}
        options = {rmap[k]: v for k, v in slurm_config.items()}
        worker_options.apply_worker_options(options)
    return {"loglevel": "debug"}


@pytest.fixture(scope="session")
def celery_worker_pool():
    if os.name == "nt":
        # "prefork" nor "process" works on windows
        return "solo"
    elif _use_slurm_pool():
        return "slurm"
    else:
        return "process"


def _use_slurm_pool() -> bool:
    return gevent_patched()


def gevent_patched() -> bool:
    try:
        from gevent.monkey import is_anything_patched
    except ImportError:
        return False

    return is_anything_patched()


@pytest.fixture()
def skip_if_gevent():
    if gevent_patched():
        pytest.skip("not supported with gevent yet")


@pytest.fixture()
def ewoks_worker(celery_session_worker, celery_worker_pool):
    yield celery_session_worker
    if celery_worker_pool == "solo":
        events.cleanup()


@pytest.fixture(scope="session")
def local_ewoks_worker(slurm_config):
    kw = {"max_workers": 8}
    if _use_slurm_pool():
        pool_type = "slurm"
        kw.update(slurm_config)
    else:
        pool_type = None
    with local.pool_context(pool_type=pool_type, **kw) as pool:
        yield
        while gc.collect():
            pass
        assert len(pool._tasks) == 0


@pytest.fixture()
def sqlite3_ewoks_events(tmpdir):
    uri = f"file:{tmpdir / 'ewoks_events.db'}"
    handlers = [
        {
            "class": "ewokscore.events.handlers.Sqlite3EwoksEventHandler",
            "arguments": [{"name": "uri", "value": uri}],
        }
    ]
    reader = instantiate_reader(uri)
    yield handlers, reader
    reader.close()
    events.cleanup()


@pytest.fixture()
def redis_ewoks_events(redisdb):
    url = f"unix://{redisdb.connection_pool.connection_kwargs['path']}"
    handlers = [
        {
            "class": "ewoksjob.events.handlers.RedisEwoksEventHandler",
            "arguments": [
                {"name": "url", "value": url},
                {"name": "ttl", "value": 3600},
            ],
        }
    ]
    reader = instantiate_reader(url)
    yield handlers, reader

    connection = redis.Redis.from_url(url)
    for key in connection.keys("ewoks:*"):
        assert connection.ttl(key) >= 0, key

    reader.close()
    events.cleanup()



"""Client side pool managed in the current process
"""

from concurrent.futures import CancelledError  # noqa F401

CancelledErrors = (CancelledError,)
from .tasks import *  # noqa F403
from .utils import *  # noqa F403
from .pool import *  # noqa F403
from .tasks import execute_graph as submit  # noqa F401
from .tasks import execute_test_graph as submit_test  # noqa F401

try:
    from gevent.monkey import is_anything_patched
    from gevent.monkey import is_module_patched
except ImportError:
    pass
else:
    if is_anything_patched() and not is_module_patched("threading"):
        raise RuntimeError("gevent patching needs to include 'threading'")



import os
import time
import signal
import logging
from threading import Thread
from multiprocessing import Manager
from multiprocessing import get_context
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import wait
from concurrent.futures import Future
from concurrent.futures.process import BrokenProcessPool
from celery.concurrency import base
from celery.concurrency import prefork
from celery.exceptions import Terminated
from billiard.einfo import ExceptionInfo
from billiard.common import reset_signals


logger = logging.getLogger(__name__)


class TaskPool(base.BasePool):
    """Multiprocessing Pool of non-daemonic processes which
    can create child processes without the need for billiard.
    """

    EXECUTOR_OPTIONS = dict()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._executor = None
        self._manager = Manager()

    def _ensure_executor(self):
        if self._executor is None:
            logger.info("Start executor ...")
            mp_context = get_context(self.EXECUTOR_OPTIONS.get("context"))
            initargs = self.options["initargs"]
            self._executor = ProcessPoolExecutor(
                max_workers=self.limit,
                initializer=process_initializer,
                initargs=initargs,
                mp_context=mp_context,
            )
            if self.EXECUTOR_OPTIONS.get("precreate"):
                timeout = self.EXECUTOR_OPTIONS.get("timeout", 10)
                logger.info("Pre-create workers ...")
                try:
                    self._executor.submit(os.getpid).result(timeout=timeout)
                except TimeoutError:
                    logger.error(f"Cannot pre-create workers within {timeout} seconds")
                    raise
            logger.info("Executor started")
        return self._executor

    def _ensure_no_executor(self):
        """Running jobs are not interrupted"""
        if self._executor is None:
            return
        logger.info("Stop executor (wait indefinitely for jobs to finish)...")
        try:
            self._executor.shutdown(wait=True)
        except BaseException:
            logger.error("Stop executor cancelled")
            raise
        logger.info("Executor stopped")
        self._executor = None

    def on_start(self):
        self._ensure_executor()

    def did_start_ok(self):
        return self._executor is not None

    def on_stop(self):
        """Called on the first CTRL-C of the worker process (warm shutdown)"""
        self._ensure_no_executor()

    def on_terminate(self):
        """Called when?"""
        self._ensure_no_executor()

    def on_close(self):
        """Called when?"""

    def terminate_job(self, pid, sig=None):
        """Called when revoking a job"""
        if sig is None:
            sig = signal.SIGTERM
        try:
            os.kill(pid, sig)
        except ProcessLookupError:
            pass

    def restart(self):
        self._ensure_no_executor()
        self._ensure_executor()

    def shrink(self, n=1):
        self._ensure_no_executor()
        self.limit = max(self.limit - n, 0)
        self._ensure_executor()

    def grow(self, n=1):
        self._ensure_no_executor()
        self.limit = max(self.limit + n, 0)
        self._ensure_executor()

    def on_apply(
        self,
        target,
        args=None,
        kwargs=None,
        accept_callback=None,
        callback=None,
        error_callback=None,
        **_,
    ):
        f = self._safe_submit(target, args, kwargs, accept_callback)

        if callback is not None or error_callback is not None:

            def done_callback(f):
                try:
                    result = f.result()
                except BrokenProcessPool:
                    # Child process was killed
                    # TODO: prefork.process_destructor(pid, exitcode)
                    if error_callback is None:
                        return
                    error_callback(parse_exception(Terminated(signal.SIGKILL)))
                except BaseException as e:
                    # Terminated: Job was cancelled
                    if error_callback is None:
                        return
                    error_callback(parse_exception(e))
                else:
                    # Job succeeded or failed: somehow the result is already
                    # converted to a tuple (failed, stringified result, time)
                    if callback is not None:
                        callback(result)

        f.add_done_callback(done_callback)
        return ApplyResult(f)

    def _safe_submit(self, target, args, kwargs, accept_callback) -> Future:
        while True:
            try:
                return self._submit(target, args, kwargs, accept_callback)
            except BrokenProcessPool:
                logger.warning("Restart executor because a child received SIGKILL")
                self.restart()

    def _submit(self, target, args, kwargs, accept_callback) -> Future:
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = dict()

        executor = self._ensure_executor()
        if accept_callback is None:
            return executor.submit(target, *args, **kwargs)

        queue = self._manager.Queue()
        t = Thread(target=accept_callback_main, args=(queue, accept_callback))
        t.daemon = True
        t.start()
        return executor.submit(subprocess_main, queue, target, args, kwargs)

    def _get_info(self):
        if self._executor is None:
            return {
                "max-concurrency": self.limit,
                "processes": [],
                "max-tasks-per-child": "N/A",
            }
        max_tasks_per_child = getattr(self._executor, "_max_tasks_per_child", None)
        return {
            "max-concurrency": self.limit,
            "processes": list(self._executor._processes),
            "max-tasks-per-child": max_tasks_per_child or "N/A",
        }

    @property
    def num_processes(self):
        if self._executor is None:
            return 0
        return len(self._executor._processes)


def process_initializer(*args):
    os.environ["FORKED_BY_MULTIPROCESSING"] = "1"
    prefork.process_initializer(*args)
    reset_signals(full=True)
    try:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    except AttributeError:
        pass


def subprocess_main(queue, target, args, kwargs):
    """Main function called in a child process"""
    try:
        queue.put(os.getpid())
        return target(*args, **kwargs)
    except SystemExit as e:
        raise Terminated(e)


def accept_callback_main(queue, accept_callback):
    """Background task that wait for the start of job execution in a child process"""
    try:
        pid = queue.get()
    except EOFError:
        return
    accept_callback(pid, time.monotonic())


def parse_exception(e: BaseException) -> ExceptionInfo:
    """Prepare an exception for the error callback"""
    try:
        raise e from None
    except type(e):
        return ExceptionInfo()


class ApplyResult:
    def __init__(self, future):
        self._future = future

    def get(self):
        return self._future.result()

    def wait(self, timeout=None):
        wait([self._future], timeout)

    def terminate(self, signum):
        raise NotImplementedError



from celery.execute import send_task
from celery.result import AsyncResult
from ..dummy_workflow import dummy_workflow

__all__ = [
    "execute_graph",
    "execute_test_graph",
    "convert_workflow",
    "discover_tasks_from_modules",
    "discover_all_tasks",
]


def execute_graph(**kw) -> AsyncResult:
    return send_task("ewoksjob.apps.ewoks.execute_graph", **kw)


def execute_test_graph(
    seconds=0, filename=None, args=None, kwargs=None, **kw
) -> AsyncResult:
    if args:
        raise TypeError("execute_test_graph does not take position arguments")
    args = (dummy_workflow(),)
    if kwargs is None:
        kwargs = dict()
    kwargs["inputs"] = [
        {"id": "sleep", "name": 0, "value": seconds},
        {"id": "result", "name": "filename", "value": filename},
    ]
    return execute_graph(args=args, kwargs=kwargs, **kw)


def convert_workflow(**kw) -> AsyncResult:
    return send_task("ewoksjob.apps.ewoks.convert_graph", **kw)


def discover_tasks_from_modules(**kw) -> AsyncResult:
    return send_task("ewoksjob.apps.ewoks.discover_tasks_from_modules", **kw)


def discover_all_tasks(**kw) -> AsyncResult:
    return send_task("ewoksjob.apps.ewoks.discover_all_tasks", **kw)



import os
import signal
from celery.concurrency import prefork
from billiard.common import reset_signals


def process_initializer(*args):
    os.environ["FORKED_BY_MULTIPROCESSING"] = "1"
    prefork.process_initializer(*args)
    reset_signals(full=True)
    try:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    except AttributeError:
        pass


process_destructor = prefork.process_destructor



import traceback
from contextlib import contextmanager

_CLIENT_EXCEPTION_MODULES = (
    "builtins",
    "concurrent",
    "celery",
    "billiard",
    "multiprocessing",
    "kombu",
    "ewoksjob",
)


class EwoksJobException(Exception):
    pass


class EwoksJobExceptionCause(Exception):
    pass


class EwoksJobBaseException(BaseException):
    pass


class EwoksJobBaseExceptionCause(BaseException):
    pass


@contextmanager
def replace_exception_for_client():
    try:
        yield
    except Exception as e:
        if _client_can_deserialize(e):
            raise
        tb = traceback.format_exc().rstrip()
        cause = EwoksJobExceptionCause(f"\n\n{tb}")
        raise EwoksJobException(str(e)) from cause
    except BaseException as e:
        if _client_can_deserialize(e):
            raise
        tb = traceback.format_exc().rstrip()
        cause = EwoksJobBaseExceptionCause(f"\n\n{tb}")
        raise EwoksJobBaseException(str(e)) from cause


def _client_can_deserialize(exception: BaseException) -> bool:
    """Returns `True` when the client-side will be able to deserialize the Exception object."""
    queue = [exception]
    while queue:
        e = queue.pop()
        emodule = e.__class__.__module__
        client_can_deserialize = any(
            emodule.startswith(mod) for mod in _CLIENT_EXCEPTION_MODULES
        )
        if not client_can_deserialize:
            return False
        if isinstance(e.__cause__, BaseException):
            queue.append(e.__cause__)
        if isinstance(e.__context__, BaseException):
            queue.append(e.__context__)
    return True



import os
from typing import Callable, Mapping, Optional, Tuple
from concurrent.futures import Future

import ewoks
from ewokscore import task_discovery

from .pool import get_active_pool
from ..dummy_workflow import dummy_workflow


__all__ = [
    "execute_graph",
    "execute_test_graph",
    "convert_workflow",
    "discover_tasks_from_modules",
    "discover_all_tasks",
]


def execute_graph(
    args: Optional[Tuple] = tuple(), kwargs: Optional[Mapping] = None
) -> Future:
    return _submit_with_jobid(ewoks.execute_graph, args=args, kwargs=kwargs)


def execute_test_graph(
    seconds=0, filename=None, kwargs: Optional[Mapping] = None
) -> Future:
    args = (dummy_workflow(),)
    if kwargs is None:
        kwargs = dict()
    kwargs["inputs"] = [
        {"id": "sleep", "name": 0, "value": seconds},
        {"id": "result", "name": "filename", "value": filename},
    ]
    return execute_graph(args=args, kwargs=kwargs)


def convert_workflow(
    args: Optional[Tuple] = tuple(), kwargs: Optional[Mapping] = None
) -> Future:
    pool = get_active_pool()
    return pool.submit(ewoks.convert_graph, args=args, kwargs=kwargs)


def discover_tasks_from_modules(
    args: Optional[Tuple] = tuple(), kwargs: Optional[Mapping] = None
) -> Future:
    pool = get_active_pool()
    return pool.submit(
        task_discovery.discover_tasks_from_modules, args=args, kwargs=kwargs
    )


def discover_all_tasks(
    args: Optional[Tuple] = tuple(), kwargs: Optional[Mapping] = None
) -> Future:
    pool = get_active_pool()
    return pool.submit(task_discovery.discover_all_tasks, args=args, kwargs=kwargs)


def _submit_with_jobid(
    func: Callable, args: Optional[Tuple] = tuple(), kwargs: Optional[Mapping] = None
) -> Future:
    pool = get_active_pool()
    if kwargs is None:
        kwargs = dict()
    execinfo = kwargs.setdefault("execinfo", dict())
    if not execinfo.get("job_id"):
        job_id = os.environ.get("SLURM_JOB_ID", None)
        if job_id:
            execinfo["job_id"] = job_id
    task_id = pool.generate_task_id(execinfo.get("job_id"))
    execinfo["job_id"] = task_id
    return pool.submit(func, task_id=task_id, args=args, kwargs=kwargs)



import time
from typing import Dict, Iterable, Optional, Tuple
import json
from datetime import datetime
from threading import Event
from ewoksutils.datetime_utils import fromisoformat

try:
    from ewokscore.variable import Variable, VariableContainer
except ImportError:
    Variable = VariableContainer = None

__all__ = ["EventType", "EwoksEventReader"]

EventType = Dict[str, str]


class EwoksEventReader:
    """Base class for receiving ewoks events on the client side."""

    def __del__(self):
        self.close()

    def close(self):
        pass

    def wait_events(
        self, timeout=None, stop_event: Optional[Event] = None, **filters
    ) -> Iterable[EventType]:
        """Yield events matching the filter until timeout is reached."""
        raise NotImplementedError

    def poll_events(
        self, timeout=None, stop_event: Optional[Event] = None, interval=0.1, **filters
    ) -> Iterable[EventType]:
        """Yield events matching the filter until timeout is reached."""
        start = time.time()
        n = 0
        while True:
            try:
                events = list(self.get_events(**filters))
            except Exception as e:
                if "no such table" not in str(e):
                    raise
            else:
                events = events[n:]
                n += len(events)
                yield from events
            if timeout is not None and (time.time() - start) > timeout:
                return
            if stop_event is not None and stop_event.is_set():
                return
            time.sleep(interval)

    def get_events(self, **filters) -> Iterable[EventType]:
        """Returns all currently available events matching the filter."""
        raise NotImplementedError

    def wait_events_with_variables(self, *args, **kwargs) -> Iterable[EventType]:
        """`get_events` with URI dereferencing."""
        for event in self.wait_events(*args, **kwargs):
            self.dereference_data_uris(event)
            yield event

    def poll_events_with_variables(self, *args, **kwargs) -> Iterable[EventType]:
        """`get_events` with URI dereferencing."""
        for event in self.poll_events(*args, **kwargs):
            self.dereference_data_uris(event)
            yield event

    def get_events_with_variables(self, *args, **kwargs) -> Iterable[EventType]:
        """`get_events` with URI dereferencing."""
        for event in self.get_events(*args, **kwargs):
            self.dereference_data_uris(event)
            yield event

    def get_full_job_events(self, **filters) -> Iterable[Tuple[EventType]]:
        """Returns events grouped by "job_id". When one event matches the filter,
        all events with the "job_id" are returned.
        """
        job_id = None
        for event in self.get_events(**filters):
            if job_id != event["job_id"]:
                job_id = event["job_id"]
                yield tuple(self.get_events(job_id=job_id))

    def get_full_job_events_with_variables(
        self, **filters
    ) -> Iterable[Tuple[EventType]]:
        """`get_full_job_events` with URI dereferencing."""
        job_id = None
        for event in self.get_events(**filters):
            if job_id != event["job_id"]:
                job_id = event["job_id"]
                yield tuple(self.get_events_with_variables(job_id=job_id))

    @staticmethod
    def dereference_data_uris(event: EventType) -> None:
        if Variable is None:
            raise ImportError("requires 'ewoks'")
        input_uris = event.get("input_uris")
        if input_uris:
            if isinstance(input_uris, str):
                input_uris = json.loads(input_uris)
            inputs = {
                uri["name"]: (
                    Variable(data_uri=uri["value"]) if uri["value"] else Variable()
                )
                for uri in input_uris
            }
            event["inputs"] = VariableContainer(inputs)
        task_uri = event.get("task_uri")
        if task_uri:
            event["outputs"] = VariableContainer(data_uri=task_uri)

    @staticmethod
    def event_passes_filter(
        event: EventType,
        starttime: Optional[datetime] = None,
        endtime: Optional[datetime] = None,
    ) -> bool:
        if not (starttime or endtime):
            return True
        time = fromisoformat(event["time"])
        if starttime is not None:
            if isinstance(starttime, str):
                starttime = fromisoformat(starttime)
            if time < starttime:
                return False
        if endtime is not None:
            if isinstance(endtime, str):
                endtime = fromisoformat(endtime)
            if time > endtime:
                return False
        return True

    @staticmethod
    def split_filter(
        starttime: Optional[datetime] = None,
        endtime: Optional[datetime] = None,
        **is_equal_filter
    ) -> Tuple[dict, dict]:
        """Splits the filter

        to be applied on the list of events fetched from the database
        """
        if starttime and not isinstance(starttime, datetime):
            raise TypeError("starttime needs to be a datetime object")
        if endtime and not isinstance(endtime, datetime):
            raise TypeError("starttime needs to be a datetime object")
        post_filter = {"starttime": starttime, "endtime": endtime}
        return is_equal_filter, post_filter



from typing import Callable, Optional


ExecutorType = Optional[Callable]
_GET_EXECUTOR: Optional[Callable[[], ExecutorType]] = None


def set_executor_getter(get_executor: Callable[[], ExecutorType]) -> None:
    """Worker pools that need to wrap their tasks can implement a
    `get_executor` function and register it here.
    """
    global _GET_EXECUTOR
    _GET_EXECUTOR = get_executor


def get_executor() -> ExecutorType:
    if _GET_EXECUTOR is None:
        return
    return _GET_EXECUTOR()



import os
import json
import socket
import logging
import redis
from typing import Dict, Optional, Tuple

from ewoksutils.logging_utils.connection import ConnectionHandler
from ewokscore.events.handlers import EwoksEventHandlerMixIn


RedisRecordType = Tuple[str, Dict[str, str]]


class RedisEwoksEventHandler(EwoksEventHandlerMixIn, ConnectionHandler):
    # TODO: https://redisql.redbeardlab.com/blog/python/using-redisql-with-python/

    def __init__(self, url: str, ttl=None):
        """An example url is "redis://localhost:10003?db=2"."""
        self._redis_url = url
        self._ttl = ttl
        super().__init__()

    def _connect(self, timeout=1) -> None:
        """This is called when no connection exists."""
        client_name = f"ewoks:writer:{socket.gethostname()}:{os.getpid()}"
        self._connection = redis.Redis.from_url(
            self._redis_url, client_name=client_name
        )

    def _disconnect(self) -> None:
        """This is called when a connection exists and is connected."""
        self._connection.close()

    def _serialize_record(self, record: logging.LogRecord) -> Optional[RedisRecordType]:
        """Convert a record to something that can be given to the connection."""
        job_id = getattr(record, "job_id", None)
        adict = dict()
        for field in self.FIELD_TYPES:
            value = getattr(record, field, None)
            adict[field] = json.dumps(value)
        return job_id, adict

    def _send_serialized_record(self, srecord: Optional[RedisRecordType]):
        """Send the output from `_serialize_record` to the connection."""
        if not srecord:
            return
        job_id, value = srecord
        n = self._connection.incrby("ewoks_events_count")
        key = f"ewoks:{job_id}:{n}"
        self._connection.hset(key, mapping=value)
        if self._ttl:
            self._connection.expire(key, self._ttl)



import pytest
import datetime
import threading
from ewokscore import events
from .utils import has_redis


@pytest.mark.skipif(not has_redis(), reason="redis-server not installed")
def test_redis(redis_ewoks_events):
    handlers, reader = redis_ewoks_events
    assert_event_reader(handlers, reader)


def test_sqlite3(sqlite3_ewoks_events):
    handlers, reader = sqlite3_ewoks_events
    assert_event_reader(handlers, reader)


@pytest.mark.skipif(not has_redis(), reason="redis-server not installed")
def test_redis_stop_wait_events(redis_ewoks_events):
    _, reader = redis_ewoks_events
    assert_stop_event(reader)


def test_sqlite3_stop_wait_events(sqlite3_ewoks_events):
    _, reader = sqlite3_ewoks_events
    assert_stop_event(reader)


def assert_event_reader(handlers, reader):
    execinfo = {
        "job_id": "123",
        "workflow_id": "456",
        "host_name": None,
        "user_name": None,
        "process_id": None,
        "handlers": handlers,
    }
    events.send_workflow_event(execinfo=execinfo, event="start")
    events.send_workflow_event(execinfo=execinfo, event="end")

    evts = list(reader.wait_events(timeout=1))
    assert len(evts) == 2

    evts = list(reader.get_events(type="end"))
    assert len(evts) == 1
    evts = list(reader.get_full_job_events(type="end"))
    assert len(evts) == 1
    assert len(evts[0]) == 2
    evts = list(reader.get_events(type="progress"))
    assert len(evts) == 0
    evts = list(reader.get_full_job_events(type="progress"))
    assert len(evts) == 0

    evts = list(reader.get_events(job_id="123"))
    assert len(evts) == 2
    evts = list(reader.get_full_job_events(job_id="123"))
    assert len(evts) == 1
    assert len(evts[0]) == 2

    now = datetime.datetime.now().astimezone()
    starttime = now - datetime.timedelta(minutes=1)
    endtime = now + datetime.timedelta(minutes=1)
    evts = list(reader.get_events(starttime=starttime, endtime=endtime))
    assert len(evts) == 2
    evts = list(
        reader.get_full_job_events(type="end", starttime=starttime, endtime=endtime)
    )
    assert len(evts) == 1
    assert len(evts[0]) == 2

    evts = list(reader.get_events(endtime=starttime))
    assert len(evts) == 0
    evts = list(reader.get_full_job_events(endtime=starttime))
    assert len(evts) == 0


def assert_stop_event(reader):
    stop_event = threading.Event()

    thread = threading.Thread(
        target=reader.wait_events, kwargs={"stop_event": stop_event}, daemon=True
    )
    thread.start()

    stop_event.set()
    thread.join(timeout=3)
    assert not thread.is_alive()



import os
import time
import logging
from types import ModuleType

logger = logging.getLogger(__name__)


def has_redis() -> bool:
    try:
        import redis  # noqa F401
    except ImportError:
        return False
    with os.popen("redis-server --version") as output:
        return bool(output.read())


def get_result(future, **kwargs):
    if hasattr(future, "get"):
        return future.get(**kwargs)
    else:
        return future.result(**kwargs)


def wait_not_finished(mod: ModuleType, expected_task_ids: set, timeout=3):
    """Wait until all running task ID's are `expected_task_ids`"""
    if mod.__name__.endswith("celery") and not has_redis():
        time.sleep(0.1)
        logger.warning("memory and sqlite do not support task monitoring")
        return
    t0 = time.time()
    while True:
        task_ids = set(mod.get_not_finished_task_ids())
        if task_ids == expected_task_ids:
            return
        dt = time.time() - t0
        if dt > timeout:
            assert task_ids == expected_task_ids, task_ids
        time.sleep(0.2)
